/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/stat.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "aslc.h"
#include "aslcsymtree.h"

#define	GEN_SDT_HEADER_SIZE	36
#define	GEN_SDT_CHECKSUM_OFFSET	9
#define	GEN_SDT_CREATOR_ID	"MSFT"
#define	GEN_SDT_CREATOR_REV	0x0100000b

#define GEN_GENERALTERM_ENCODING(ptr)	do {				\
	Byte	byte_encoding;						\
	byte_encoding = ptr->symbol.common.encoding;			\
	bcopy(&byte_encoding, gen_amlcode_ptr, 1);			\
	gen_amlcode_ptr += 1;						\
} while (0)

#define GEN_INTEGER_ENCODING(val)	do {				\
	if (val <= 0xff) {						\
		Byte	byte_encoding;					\
		byte_encoding = val & 0xff;				\
		bcopy(&byte_encoding, gen_amlcode_ptr, 1);		\
		gen_amlcode_ptr += 1;					\
	} else if (val <= 0xffff) {					\
		Word	word_encoding;					\
		word_encoding = val & 0xffff;				\
		bcopy(&word_encoding, gen_amlcode_ptr, 2);		\
		gen_amlcode_ptr += 2;					\
	} else if (val <= 0xffffffff) {					\
		DWord	dword_encoding;					\
		dword_encoding = val & 0xffffffff;			\
		bcopy(&dword_encoding, gen_amlcode_ptr, 4);		\
		gen_amlcode_ptr += 4;					\
	} else {							\
		QWord	qword_encoding;					\
		qword_encoding = val;					\
		bcopy(&qword_encoding, gen_amlcode_ptr, 8);		\
		gen_amlcode_ptr += 8;					\
	}								\
} while (0)

#define GEN_INTEGER_PREFIX(val)	do {					\
	Byte	byte_prefix;						\
	if (val <= 0xff) {						\
		byte_prefix = 0x0a;					\
	} else if (val <= 0xffff) {					\
		byte_prefix = 0x0b;					\
	} else if (val <= 0xffffffff) {					\
		byte_prefix = 0x0c;					\
	} else {							\
		byte_prefix = 0x0e;					\
	}								\
	bcopy(&byte_prefix, gen_amlcode_ptr, 1);			\
	gen_amlcode_ptr += 1;						\
} while (0)

static u_int8_t		*gen_amlcode;
static u_int8_t		*gen_amlcode_ptr;
static u_int32_t	 gen_amlcode_len;

static int
generator_pkglength_encoding(DWord pkglength, Byte *encoding)
{
	int	len;
	Byte	code[4];

	len = 0;
	bzero(code, sizeof(code));
	if (pkglength <= 0x0000003f) {	/* 6bit */
		len = 1;
		code[0] = pkglength;
		goto out;
	}
	if (pkglength <= 0x00000fff) {	/* 12bit */
		len = 2;
		code[0] = (1 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		goto out;
	}
	if (pkglength <= 0x000fffff) {	/* 20bit */
		len = 3;
		code[0] = (2 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		code[2] = (pkglength >> 12) & 0xff;
		goto out;
	}
	if (pkglength <= 0x0fffffff) {	/* 28bit */
		len = 4;
		code[0] = (3 << 6) + (pkglength & 0xf);
		code[1] = (pkglength >> 4) & 0xff;
		code[2] = (pkglength >> 12) & 0xff;
		code[3] = (pkglength >> 20) & 0xff;
		goto out;
	}
out:
	if (encoding != NULL) {
		bcopy(code, encoding, sizeof(code));
	}
	return(len);
}

static void
EncodeDefinitionBlock(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = GEN_SDT_HEADER_SIZE;
	ptr->symbol.DefinitionBlock.CreatorID = strdup(GEN_SDT_CREATOR_ID);
	ptr->symbol.DefinitionBlock.CreatorRevision = GEN_SDT_CREATOR_REV;
}

static void
GenerateDefinitionBlock(struct asl_symtree *ptr)
{
	Byte		byte;
	DWord		dword;

	bcopy(ptr->symbol.DefinitionBlock.TableSignature, gen_amlcode_ptr, 4);
	gen_amlcode_ptr += 4;

	dword = ptr->symbol.common.pkglength;
	bcopy(&dword, gen_amlcode_ptr, 4);
	gen_amlcode_ptr += 4;

	byte = ptr->symbol.DefinitionBlock.ComplianceRevision;
	bcopy(&byte, gen_amlcode_ptr, 1);
	gen_amlcode_ptr += 1;

	byte = 0;		/* Checksum, should be adjusted later */
	bcopy(&byte, gen_amlcode_ptr, 1);
	gen_amlcode_ptr += 1;

	strncpy((char *)gen_amlcode_ptr, ptr->symbol.DefinitionBlock.OEMID, 6);
	gen_amlcode_ptr += 6;

	strncpy((char *)gen_amlcode_ptr, ptr->symbol.DefinitionBlock.TableID, 8);
	gen_amlcode_ptr += 8;

	dword = ptr->symbol.DefinitionBlock.OEMRevision;
	bcopy(&dword, gen_amlcode_ptr, 4);
	gen_amlcode_ptr += 4;

	strncpy((char *)gen_amlcode_ptr, ptr->symbol.DefinitionBlock.CreatorID, 4);
	gen_amlcode_ptr += 4;

	dword = ptr->symbol.DefinitionBlock.CreatorRevision;
	bcopy(&dword, gen_amlcode_ptr, 4);
	gen_amlcode_ptr += 4;
}

static void
EncodeCreateByteFieldTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x8c;
}

static void
GenerateCreateByteFieldTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeMethodTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x14;
	ptr->symbol.common.pkglength += strlen(ptr->symbol.Method.MethodName);
	ptr->symbol.common.pkglength += 1;
	ptr->symbol.Method.Flags = (ptr->symbol.Method.NumArgs |
				    ptr->symbol.Method.SerializeRule << 3);
}

static void
GenerateMethodTerm(struct asl_symtree *ptr)
{
	Byte	byte;
	Byte	encoding[4];
	size_t	len;

	GEN_GENERALTERM_ENCODING(ptr);

	len = generator_pkglength_encoding(ptr->symbol.common.pkglength, encoding);
	bcopy(encoding, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;

	len = strlen(ptr->symbol.Method.MethodName);
	bcopy(ptr->symbol.Method.MethodName, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;

	byte = ptr->symbol.Method.Flags;
	bcopy(&byte, gen_amlcode_ptr, 1);
	gen_amlcode_ptr += 1;
}

static void
EncodeWhileTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0xa2;
}

static void
GenerateWhileTerm(struct asl_symtree *ptr)
{
	Byte	encoding[4];
	size_t	len;

	GEN_GENERALTERM_ENCODING(ptr);
	len = generator_pkglength_encoding(ptr->symbol.common.pkglength, encoding);
	bcopy(encoding, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;
}

static void
EncodeDecTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x76;
}

static void
GenerateDecTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeRefOfTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x71;
}

static void
GenerateRefOfTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeStoreTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x70;
	ptr->child->symbol.common.prefix++;	/* prefix is expected for source */
}

static void
GenerateStoreTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeArgTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x68 + ptr->symbol.common.val;
}

static void
GenerateArgTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeLocalTerm(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x60 + ptr->symbol.common.val;
}

static void
GenerateLocalTerm(struct asl_symtree *ptr)
{
	GEN_GENERALTERM_ENCODING(ptr);
}

static void
EncodeByteConst(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 1;
	if (ptr->symbol.common.prefix) {
		ptr->symbol.common.pkglength++;
		ptr->symbol.common.prefix = 0x0a;
	}
}

static void
GenerateByteConst(struct asl_symtree *ptr)
{
	Byte	byte;

	if (ptr->symbol.common.prefix) {
		byte = ptr->symbol.common.prefix;
		bcopy(&byte, gen_amlcode_ptr, 1);
		gen_amlcode_ptr += 1;
	}

	byte = (ptr->symbol.common.val & 0xff);
	bcopy(&byte, gen_amlcode_ptr, 1);
	gen_amlcode_ptr += 1;
}

static void
EncodeWordConst(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 2;
	if (ptr->symbol.common.prefix) {
		ptr->symbol.common.pkglength++;
		ptr->symbol.common.prefix = 0x0b;
	}
}

static void
GenerateWordConst(struct asl_symtree *ptr)
{
	Byte	byte;
	Word	word;

	if (ptr->symbol.common.prefix) {
		byte = ptr->symbol.common.prefix;
		bcopy(&byte, gen_amlcode_ptr, 1);
		gen_amlcode_ptr += 1;
	}

	word = ptr->symbol.common.val & 0xffff;
	bcopy(&word, gen_amlcode_ptr, 2);
	gen_amlcode_ptr += 2;
}

static void
EncodeDWordConst(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 4;
	if (ptr->symbol.common.prefix) {
		ptr->symbol.common.pkglength++;
		ptr->symbol.common.prefix = 0x0c;
	}
}

static void
GenerateDWordConst(struct asl_symtree *ptr)
{
	Byte	byte;
	DWord	dword;

	if (ptr->symbol.common.prefix) {
		byte = ptr->symbol.common.prefix;
		bcopy(&byte, gen_amlcode_ptr, 1);
		gen_amlcode_ptr += 1;
	}

	dword = ptr->symbol.common.val & 0xffffffff;
	bcopy(&dword, gen_amlcode_ptr, 4);
	gen_amlcode_ptr += 4;
}

static void
EncodeQWordConst(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = 8;
	if (ptr->symbol.common.prefix) {
		ptr->symbol.common.pkglength++;
		ptr->symbol.common.prefix = 0x0e;	/* XXX */
	}
}

static void
GenerateQWordConst(struct asl_symtree *ptr)
{
	Byte	byte;
	DWord	dword;

	if (ptr->symbol.common.prefix) {
		byte = ptr->symbol.common.prefix;
		bcopy(&byte, gen_amlcode_ptr, 1);
		gen_amlcode_ptr += 1;
	}

	dword = ptr->symbol.common.val;
	bcopy(&dword, gen_amlcode_ptr, 8);
	gen_amlcode_ptr += 8;
}

static void
EncodeString(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = strlen(ptr->symbol.String.ptr) + 1;
	if (ptr->symbol.common.prefix) {
		ptr->symbol.common.pkglength++;
		ptr->symbol.common.prefix = 0x0d;
	}
}

static void
GenerateString(struct asl_symtree *ptr)
{
	Byte	byte;
	size_t	len;

	byte = ptr->symbol.common.prefix;
	bcopy(&byte, gen_amlcode_ptr, 1);
	gen_amlcode_ptr += 1;

	len = strlen(ptr->symbol.String.ptr) + 1;
	bcopy(ptr->symbol.String.ptr, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;
}

static void
EncodeBufferTerm(struct asl_symtree *ptr)
{
	int	size;
	struct asl_symtree *block;

	size = 0;
	
	ptr->symbol.common.pkglength = 1;
	ptr->symbol.common.encoding = 0x11;
	if (ptr->child->symbol.common.type == SYMTYPE_BufferBlock) {
		/* buffer size is unknown */
		block = ptr->child->child;
		if (block) {
			if (block->symbol.common.type == SYMTYPE_String) {
				size = strlen(block->symbol.String.ptr) + 1;
				block->symbol.common.prefix++;
			} else {
				while (block) {
					size++;
					block = block->brother;
				}
			}
		}
		ptr->symbol.Buffer.BufferSize = size;
		while (size) {
			ptr->symbol.common.pkglength++;
			size = size >> 8;
		}
		ptr->symbol.common.pkglength++;	/* for prefix of BufferSize */
	} else {
		ptr->child->symbol.common.prefix++;
	}
}

static void
GenerateBufferTerm(struct asl_symtree *ptr)
{
	Byte	encoding[4];
	size_t	len;

	GEN_GENERALTERM_ENCODING(ptr);
	len = generator_pkglength_encoding(ptr->symbol.common.pkglength, encoding);
	bcopy(encoding, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;

	if (ptr->symbol.Buffer.BufferSize != 0) {
		GEN_INTEGER_PREFIX(ptr->symbol.Buffer.BufferSize);
		GEN_INTEGER_ENCODING(ptr->symbol.Buffer.BufferSize);
	}
}

static void
EncodeNameString(struct asl_symtree *ptr)
{
	ptr->symbol.common.pkglength = strlen(ptr->symbol.NameString.ptr);
}

static void
GenerateNameString(struct asl_symtree *ptr)
{
	size_t	len;

	len = strlen(ptr->symbol.NameString.ptr);
	bcopy(ptr->symbol.NameString.ptr, gen_amlcode_ptr, len);
	gen_amlcode_ptr += len;
}


static void
generator_encode_term(struct asl_symtree *ptr)
{
	switch (ptr->symbol.common.type) {
	case SYMTYPE_DefinitionBlock:
		EncodeDefinitionBlock(ptr);
		break;
	case SYMTYPE_TermList:
		break;
	case SYMTYPE_UserTerm:
		break;
	case SYMTYPE_CreateByteFieldTerm:
		EncodeCreateByteFieldTerm(ptr);
		break;
	case SYMTYPE_MethodTerm:
		EncodeMethodTerm(ptr);
		break;
	case SYMTYPE_WhileTerm:
		EncodeWhileTerm(ptr);
		break;
	case SYMTYPE_DecTerm:
		EncodeDecTerm(ptr);
		break;
	case SYMTYPE_RefOfTerm:
		EncodeRefOfTerm(ptr);
		break;
	case SYMTYPE_StoreTerm:
		EncodeStoreTerm(ptr);
		break;
	case SYMTYPE_ArgTerm:
		EncodeArgTerm(ptr);
		break;
	case SYMTYPE_LocalTerm:
		EncodeLocalTerm(ptr);
		break;
	case SYMTYPE_ByteConst:
		EncodeByteConst(ptr);
		break;
	case SYMTYPE_WordConst:
		EncodeWordConst(ptr);
		break;
	case SYMTYPE_DWordConst:
		EncodeDWordConst(ptr);
		break;
	case SYMTYPE_QWordConst:
		EncodeQWordConst(ptr);
		break;
	case SYMTYPE_String:
		EncodeString(ptr);
		break;
	case SYMTYPE_BufferTerm:
		EncodeBufferTerm(ptr);
		break;
	case SYMTYPE_BufferBlock:
		break;
	case SYMTYPE_NameString:
		EncodeNameString(ptr);
		break;
	default:
		break;
	}

	if (ptr->child != NULL) {
		generator_encode_term(ptr->child);
	}

	if (ptr->brother != NULL) {
		generator_encode_term(ptr->brother);
	}
}

static void
generator_calculate_pkglength(struct asl_symtree *ptr)
{
	if (ptr->child != NULL) {
		generator_calculate_pkglength(ptr->child);
	}

	if (ptr->brother != NULL) {
		generator_calculate_pkglength(ptr->brother);
	}

	if (ptr->parent != NULL) {
		ptr->parent->symbol.common.pkglength += ptr->symbol.common.pkglength;
	}
}

static void
generator_fixup_pkglength(struct asl_symtree *ptr)
{
	int	len;
	struct asl_symtree	*parent;

	if (ptr->child != NULL) {
		generator_fixup_pkglength(ptr->child);
	}

	if (ptr->brother != NULL) {
		generator_fixup_pkglength(ptr->brother);
	}

	switch (ptr->symbol.common.type) {
	case SYMTYPE_DefinitionBlock:
		ptr->symbol.common.pkglength += ptr->brother->symbol.common.pkglength;
		break;
	case SYMTYPE_MethodTerm:
	case SYMTYPE_WhileTerm:
	case SYMTYPE_BufferTerm:
		ptr->symbol.common.pkglength--;	/* exclude the it's encoding */
		len = generator_pkglength_encoding(ptr->symbol.common.pkglength, NULL);
		ptr->symbol.common.pkglength += len;
		for (parent = ptr->parent; parent; parent = parent->parent) {
			parent->symbol.common.pkglength += len;
		}
		break;
	default:
		break;
	}
}

static void
generator_generate_amlcode(struct asl_symtree *ptr)
{
	switch (ptr->symbol.common.type) {
	case SYMTYPE_DefinitionBlock:
		GenerateDefinitionBlock(ptr);
		break;
	case SYMTYPE_TermList:
		break;
	case SYMTYPE_UserTerm:
		break;
	case SYMTYPE_CreateByteFieldTerm:
		GenerateCreateByteFieldTerm(ptr);
		break;
	case SYMTYPE_MethodTerm:
		GenerateMethodTerm(ptr);
		break;
	case SYMTYPE_WhileTerm:
		GenerateWhileTerm(ptr);
		break;
	case SYMTYPE_DecTerm:
		GenerateDecTerm(ptr);
		break;
	case SYMTYPE_RefOfTerm:
		GenerateRefOfTerm(ptr);
		break;
	case SYMTYPE_StoreTerm:
		GenerateStoreTerm(ptr);
		break;
	case SYMTYPE_ArgTerm:
		GenerateArgTerm(ptr);
		break;
	case SYMTYPE_LocalTerm:
		GenerateLocalTerm(ptr);
		break;
	case SYMTYPE_ByteConst:
		GenerateByteConst(ptr);
		break;
	case SYMTYPE_WordConst:
		GenerateWordConst(ptr);
		break;
	case SYMTYPE_DWordConst:
		GenerateDWordConst(ptr);
		break;
	case SYMTYPE_QWordConst:
		GenerateQWordConst(ptr);
		break;
	case SYMTYPE_String:
		GenerateString(ptr);
		break;
	case SYMTYPE_BufferTerm:
		GenerateBufferTerm(ptr);
		break;
	case SYMTYPE_BufferBlock:
		break;
	case SYMTYPE_NameString:
		GenerateNameString(ptr);
		break;
	default:
		break;
	}
	if (ptr->child != NULL) {
		generator_generate_amlcode(ptr->child);
	}

	if (ptr->brother != NULL) {
		generator_generate_amlcode(ptr->brother);
	}
}

static void
generator_fixup_checksum(void)
{
	Byte	cksm;
	DWord	i;
	
	cksm = 0;
	for (i = 0; i < gen_amlcode_len; i++) {
		cksm += gen_amlcode[i];
	}
	cksm = 0x100 - cksm;
	gen_amlcode[GEN_SDT_CHECKSUM_OFFSET] = cksm;
}

void
generator_dump_amlcode(void)
{
	DWord	i;

	for (i = 0; i < gen_amlcode_len; i++) {
		if (i != 0 && (i % 16) == 0) {
			printf("\n");
		}
		printf("%02x ", gen_amlcode[i]);
	}
	printf("\n");
}

int
generator_create_amlfile(const char *filename)
{
	int	fd, bytes;
	mode_t	mode;

	if (filename == NULL) {
		return(0);
	}

	mode = (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
	fd = open(filename, O_WRONLY | O_CREAT, mode);
	if (fd == -1) {
		err(1, "opening %s", filename);
		return(0);
	}

	bytes = write(fd, gen_amlcode, gen_amlcode_len);
	if (bytes != (int)gen_amlcode_len) {
		err(1, "writing %s", filename);
		return(bytes);
	}

	close(fd);
	return(bytes);
}

void
generator_main(void)
{
	struct asl_symtree	*root;

	root = symtree_getroot();
	generator_encode_term(root);
	generator_calculate_pkglength(root);
	generator_fixup_pkglength(root);
	gen_amlcode_len = root->symbol.common.pkglength;
	gen_amlcode = malloc(gen_amlcode_len);
	if (gen_amlcode == NULL) {
		err(1, "No more space for AML code generation");
	}
	bzero(gen_amlcode, gen_amlcode_len);
	gen_amlcode_ptr = gen_amlcode;
	generator_generate_amlcode(root);
	generator_fixup_checksum();
}
