%{
/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/types.h>

#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "aslc.h"
#include "y.tab.h"

#define ACTION_OBJECTTYPEKEYWORD(objtype)	do {			\
	yylval.num = objtype;						\
	return(TOK_OBJECTTYPEKEYWORD);					\
} while (0)

#define ACTION_ACCESSTYPEKEYWORD(acctype)	do {			\
	yylval.num = acctype;						\
	return(TOK_ACCESSTYPEKEYWORD);					\
} while (0)

#define ACTION_ACCESSATTRIBKEYWORD(accattrib)	do {			\
	yylval.num = accattrib;						\
	return(TOK_ACCESSATTRIBKEYWORD);				\
} while (0)

#define ACTION_LOCKRULEKEYWORD(value)		do {			\
	yylval.num = value;						\
	return(TOK_LOCKRULEKEYWORD);					\
} while (0)

#define ACTION_UPDATERULEKEYWORD(update)	do {			\
	yylval.num = update;						\
	return(TOK_UPDATERULEKEYWORD);					\
} while (0)

#define ACTION_REGIONSPACEKEYWORD(region)	do {			\
	yylval.num = region;						\
	return(TOK_REGIONSPACEKEYWORD);					\
} while (0)

#define ACTION_SERIALIZERULEKEYWORD(value)	do {			\
	yylval.num = value;						\
	return(TOK_SERIALIZERULEKEYWORD);				\
} while (0)

#define ACTION_MATCHOPKEYWORD(match)	do {				\
	yylval.num = match;						\
	return(TOK_MATCHOPKEYWORD);					\
} while (0)

#define ACTION_DMATYPEKEYWORD(dmatype)	do {				\
	yylval.num = dmatype;						\
	return(TOK_DMATYPEKEYWORD);					\
} while (0)

#define ACTION_BUSMASTERKEYWORD(busmaster)	do {			\
	yylval.num = busmaster;						\
	return(TOK_BUSMASTERKEYWORD);					\
} while (0)

#define ACTION_XFERTYPEKEYWORD(xfertype)	do {			\
	yylval.num = xfertype;						\
	return(TOK_XFERTYPEKEYWORD);					\
} while (0)

#define ACTION_RESOURCETYPEKEYWORD(resource)	do {			\
	yylval.num = resource;						\
	return(TOK_RESOURCETYPEKEYWORD);				\
} while (0)

#define ACTION_MINKEYWORD(minkeyword)	do {				\
	yylval.num = minkeyword;					\
	return(TOK_MINKEYWORD);						\
} while (0)

#define ACTION_MAXKEYWORD(maxkeyword)	do {				\
	yylval.num = maxkeyword;					\
	return(TOK_MAXKEYWORD);						\
} while (0)

#define ACTION_DECODEKEYWORD(decode)	do {				\
	yylval.num = decode;						\
	return(TOK_DECODEKEYWORD);					\
} while (0)

#define ACTION_RANGETYPEKEYWORD(range)	do {				\
	yylval.num = range;						\
	return(TOK_RANGETYPEKEYWORD);					\
} while (0)

#define ACTION_MEMTYPEKEYWORD(memtype)	do {				\
	yylval.num = memtype;						\
	return(TOK_MEMTYPEKEYWORD);					\
} while (0)

#define ACTION_READWRITEKEYWORD(rw)	do {				\
	yylval.num = rw;						\
	return(TOK_READWRITEKEYWORD);					\
} while (0)

#define ACTION_INTERRUPTTYPEKEYWORD(inttype)	do {			\
	yylval.num = inttype;						\
	return(TOK_INTERRUPTTYPEKEYWORD);				\
} while (0)

#define ACTION_INTERRUPTLEVELKEYWORD(intlevel)	do {			\
	yylval.num = intlevel;						\
	return(TOK_INTERRUPTLEVELKEYWORD);				\
} while (0)

#define ACTION_SHARETYPEKEYWORD(share)	do {				\
	yylval.num = share;						\
	return(TOK_SHARETYPEKEYWORD);					\
} while (0)

#define ACTION_IODECODEKEYWORD(iodecode)	do {			\
	yylval.num = iodecode;						\
	return(TOK_IODECODEKEYWORD);					\
} while (0)

#define ACTION_TYPEKEYWORD(type)		do {			\
	yylval.num = type;						\
	return(TOK_TYPEKEYWORD);					\
} while (0)

#define ACTION_TRANSLATIONKEYWORD(trans)	do {			\
	yylval.num = trans;						\
	return(TOK_TRANSLATIONKEYWORD);					\
} while (0)

#define ACTION_ADDRESSKEYWORD(addr)		do {			\
	yylval.num = addr;						\
	return(TOK_ADDRESSKEYWORD);					\
} while (0)

#define ACTION_VARIABLE(pos, token)	do {				\
	yylval.num = strtol(&yytext[pos], NULL, 10);			\
	return(token);							\
} while (0)

#define ACTION_INTEGER(base, bytewidth)	do {				\
	u_quad_t	val64;						\
	yylval.lit.type = LITERAL_INTEGER;				\
	val64 = strtouq(yytext, NULL, base);				\
	yylval.lit.num.val = val64 & 0xffffffff;			\
	yylval.lit.num.bytes = bytewidth;				\
	switch (bytewidth) {						\
	case 0:								\
		if (yylval.lit.num.val <= 0xff) {			\
			yylval.lit.num.bytes = 1;			\
			return(TOK_BYTECONST);				\
		}							\
		if (yylval.lit.num.val <= 0xffff) {			\
			yylval.lit.num.bytes = 2;			\
			return(TOK_WORDCONST);				\
		}							\
		if (yylval.lit.num.val <= 0xffffffff) {			\
			yylval.lit.num.bytes = 4;			\
			return(TOK_DWORDCONST);				\
		}							\
		yylval.lit.num.bytes = 8;				\
		return(TOK_QWORDCONST);					\
	case 1:								\
		return(TOK_BYTECONST);					\
	case 2:								\
		return(TOK_WORDCONST);					\
	case 4:								\
		return(TOK_DWORDCONST);					\
	case 8:								\
		return(TOK_QWORDCONST);					\
	}								\
} while (0)

#define ACTION_CONSTANT(value, bytewidth, token)	do {		\
	yylval.lit.type = LITERAL_INTEGER;				\
	yylval.lit.num.val = value;					\
	yylval.lit.num.bytes = bytewidth;				\
	return(token);							\
} while (0)

int		parser_lineno;

static int	comment_start = 1;

%}

%x CODE
%x COMMENT

%%

%{
	if (parser_not_initialized) {
		BEGIN(CODE);
		parser_lineno = 1;
		parser_not_initialized = 0;
	}
%}

 /*
  * Newline
  */

<CODE,COMMENT>\n	{ parser_lineno++; }

 /*
  * Comment
  */

 /* one-line omment */
<CODE>"//".*$		;

 /* normal style comment */
<CODE>"/*"		{ comment_start = parser_lineno; BEGIN(COMMENT); }
<COMMENT>[^*\n]*	;	/* eat anything that's not a '*' */
<COMMENT>"*"+[^*/\n]*	;	/* eat up '*'s not followed by '/'s */
<COMMENT>"*"+"/"	{ BEGIN(CODE); }
<COMMENT><<EOF>>	{
				parser_error("unterminated comment "
				    "beginning from line %d\n", comment_start);
				 yyterminate();
			}

 /*
  * ASL code
  */

<CODE>[ \t]+		;
<CODE>[,{}()]		{ return(yytext[0]); }

<CODE>DefinitionBlock	{ return(TOK_DEFINITIONBLOCK); }

<CODE>Include		{ return(TOK_INCLUDE); }
<CODE>External		{ return(TOK_EXTERNAL); }

 /* Named Object */
<CODE>BankField		{ return(TOK_BANKFIELD); }
<CODE>Offset		{ return(TOK_OFFSET); }
<CODE>AccessAs		{ return(TOK_ACCESSAS); }
<CODE>CreateBitField	{ return(TOK_CREATEBITFIELD); }
<CODE>CreateByteField	{ return(TOK_CREATEBYTEFIELD); }
<CODE>CreateDWordField	{ return(TOK_CREATEDWORDFIELD); }
<CODE>CreateField	{ return(TOK_CREATEFIELD); }
<CODE>CreateQWordField	{ return(TOK_CREATEQWORDFIELD); }
<CODE>CreateWordField	{ return(TOK_CREATEWORDFIELD); }
<CODE>DataTableRegion	{ return(TOK_DATATABLEREGION); }
<CODE>Device		{ return(TOK_DEVICE); }
<CODE>Event		{ return(TOK_EVENT); }
<CODE>Field		{ return(TOK_FIELD); }
<CODE>IndexField	{ return(TOK_INDEXFIELD); }
<CODE>Method		{ return(TOK_METHOD); }
<CODE>Mutex		{ return(TOK_MUTEX); }
<CODE>OperationRegion	{ return(TOK_OPERATIONREGION); }
<CODE>PowerResource	{ return(TOK_POWERRESOURCE); }
<CODE>Processor		{ return(TOK_PROCESSOR); }
<CODE>ThermalZone	{ return(TOK_THERMALZONE); }
<CODE>Alias		{ return(TOK_ALIAS); }
<CODE>Name		{ return(TOK_NAME); }
<CODE>Scope		{ return(TOK_SCOPE); }

 /* Type1Opcode */
<CODE>Break		{ return(TOK_BREAK); }
<CODE>BreakPoint	{ return(TOK_BREAKPOINT); }
<CODE>Continue		{ return(TOK_CONTINUE); }
<CODE>Fatal		{ return(TOK_FATAL); }
<CODE>If		{ return(TOK_IF); }
<CODE>Else		{ return(TOK_ELSE); }
<CODE>ElseIf		{ return(TOK_ELSEIF); }
<CODE>Load		{ return(TOK_LOAD); }
<CODE>Noop		{ return(TOK_NOOP); }
<CODE>Notify		{ return(TOK_NOTIFY); }
<CODE>Release		{ return(TOK_RELEASE); }
<CODE>Reset		{ return(TOK_RESET); }
<CODE>Return		{ return(TOK_RETURN); }
<CODE>Signal		{ return(TOK_SIGNAL); }
<CODE>Sleep		{ return(TOK_SLEEP); }
<CODE>Stall		{ return(TOK_STALL); }
<CODE>Switch		{ return(TOK_SWITCH); }
<CODE>Case		{ return(TOK_CASE); }
<CODE>Default		{ return(TOK_DEFAULT); }
<CODE>Unload		{ return(TOK_UNLOAD); }
<CODE>While		{ return(TOK_WHILE); }

 /* Type2Opcode */
<CODE>Acquire		{ return(TOK_ACQUIRE); }
<CODE>Add		{ return(TOK_ADD); }
<CODE>And		{ return(TOK_AND); }
<CODE>Buff		{ return(TOK_BUFF); }
<CODE>Concatenate	{ return(TOK_CONCATENATE); }
<CODE>ConcatenateResTemplate	{ return(TOK_CONCATENATERESTEMPLATE); }
<CODE>CondRefOf		{ return(TOK_CONDREFOF); }
<CODE>Copy		{ return(TOK_COPY); }
<CODE>Decrement		{ return(TOK_DECREMENT); }
<CODE>DecStr		{ return(TOK_DECSTR); }
<CODE>DerefOf		{ return(TOK_DEREFOF); }
<CODE>Divide		{ return(TOK_DIVIDE); }
<CODE>FindSetLeftBit	{ return(TOK_FINDSETLEFTBIT); }
<CODE>FindSetRightBit	{ return(TOK_FINDSETRIGHTBIT); }
<CODE>FromBCD		{ return(TOK_FROMBCD); }
<CODE>HexStr		{ return(TOK_HEXSTR); }
<CODE>Increment		{ return(TOK_INCREMENT); }
<CODE>Index		{ return(TOK_INDEX); }
<CODE>Int		{ return(TOK_INT); }
<CODE>LAnd		{ return(TOK_LAND); }
<CODE>LEqual		{ return(TOK_LEQUAL); }
<CODE>LGreater		{ return(TOK_LGREATER); }
<CODE>LGreaterEqual	{ return(TOK_LGREATEREQUAL); }
<CODE>LLess		{ return(TOK_LLESS); }
<CODE>LLessEqual	{ return(TOK_LLESSEQUAL); }
<CODE>LNot		{ return(TOK_LNOT); }
<CODE>LNotEqual		{ return(TOK_LNOTEQUAL); }
<CODE>LoadTable		{ return(TOK_LOADTABLE); }
<CODE>LOr		{ return(TOK_LOR); }
<CODE>Match		{ return(TOK_MATCH); }
<CODE>Mid		{ return(TOK_MID); }
<CODE>Mod		{ return(TOK_MOD); }
<CODE>Multiply		{ return(TOK_MULTIPLY); }
<CODE>NAnd		{ return(TOK_NAND); }
<CODE>NOr		{ return(TOK_NOR); }
<CODE>Not		{ return(TOK_NOT); }
<CODE>ObjectType	{ return(TOK_OBJECTTYPE); }
<CODE>Or		{ return(TOK_OR); }
<CODE>RefOf		{ return(TOK_REFOF); }
<CODE>ShiftLeft		{ return(TOK_SHIFTLEFT); }
<CODE>ShiftRight	{ return(TOK_SHIFTRIGHT); }
<CODE>SizeOf		{ return(TOK_SIZEOF); }
<CODE>Store		{ return(TOK_STORE); }
<CODE>String		{ return(TOK_STRING); }
<CODE>Subtract		{ return(TOK_SUBTRACT); }
<CODE>ToBCD		{ return(TOK_TOBCD); }
<CODE>Wait		{ return(TOK_WAIT); }
<CODE>XOr		{ return(TOK_XOR); }

 /* Keyword */
<CODE>UnknownObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_UNKNOWN); }
<CODE>IntObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_INT); }
<CODE>StrObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_STR); }
<CODE>BuffObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_BUFF); }
<CODE>PkgObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_PKG); }
<CODE>FieldUnitObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_FIELDUNIT); }
<CODE>DeviceObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_DEVICE); }
<CODE>EventObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_EVENT); }
<CODE>MethodObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_METHOD); }
<CODE>MutexObj		{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_MUTEX); }
<CODE>OpRegionObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_OPREGION); }
<CODE>PowerResObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_POWERRES); }
<CODE>ThermalZoneObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_THERMALZONE); }
<CODE>BuffFieldObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_BUFFERFIELD); }
<CODE>DDBHandleObj	{ ACTION_OBJECTTYPEKEYWORD(OBJTYPE_DDBHANDLE); }

<CODE>AnyAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_ANY); }
<CODE>ByteAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_BYTE); }
<CODE>WordAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_WORD); }
<CODE>DWordAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_DWORD); }
<CODE>QWordAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_QWORD); }
<CODE>BufferAcc		{ ACTION_ACCESSTYPEKEYWORD(ACCTYPE_BUFFER); }

<CODE>SMBQuick		{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBQUICK); }
<CODE>SMBSendReceive	{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBSENDRECEIVE); }
<CODE>SMBByte		{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBBYTE); }
<CODE>SMBWord		{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBWORD); }
<CODE>SMBBlock		{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBBLOCK); }
<CODE>SMBProcessCall	{ ACTION_ACCESSATTRIBKEYWORD(ACCATTRIB_SMBPROCESSCALL); }

<CODE>Lock		{ ACTION_LOCKRULEKEYWORD(1); }
<CODE>NoLock		{ ACTION_LOCKRULEKEYWORD(0); }

<CODE>Preserve		{ ACTION_UPDATERULEKEYWORD(UPDATE_PRESERVE); }
<CODE>WriteAsOnes	{ ACTION_UPDATERULEKEYWORD(UPDATE_ASONES); }
<CODE>WriteAsZeros	{ ACTION_UPDATERULEKEYWORD(UPDATE_ASZEROS); }

<CODE>UserDefRegionSpace {ACTION_REGIONSPACEKEYWORD(REGION_USERDEF); }
<CODE>SystemIO		{ ACTION_REGIONSPACEKEYWORD(REGION_SYSTEMIO); }
<CODE>SystemMemory	{ ACTION_REGIONSPACEKEYWORD(REGION_SYSTEMMEM); }
<CODE>PCI_Config	{ ACTION_REGIONSPACEKEYWORD(REGION_PCICONFIG); }
<CODE>EmbeddedControl	{ ACTION_REGIONSPACEKEYWORD(REGION_EMBCONTROL); }
<CODE>SMBus		{ ACTION_REGIONSPACEKEYWORD(REGION_SMBUS); }
<CODE>CMOS		{ ACTION_REGIONSPACEKEYWORD(REGION_CMOS); }
<CODE>PciBarTarget	{ ACTION_REGIONSPACEKEYWORD(REGION_PCIBARTARGET); }

<CODE>FFixedHW		{ return(TOK_FFIXEDHW); }

<CODE>Serialized	{ ACTION_SERIALIZERULEKEYWORD(1); }
<CODE>NotSerialized	{ ACTION_SERIALIZERULEKEYWORD(0); }

<CODE>MTR		{ ACTION_MATCHOPKEYWORD(MATCH_MTR); }
<CODE>MEQ		{ ACTION_MATCHOPKEYWORD(MATCH_MEQ); }
<CODE>MLE		{ ACTION_MATCHOPKEYWORD(MATCH_MLE); }
<CODE>MLT		{ ACTION_MATCHOPKEYWORD(MATCH_MLT); }
<CODE>MGE		{ ACTION_MATCHOPKEYWORD(MATCH_MGE); }
<CODE>MGT		{ ACTION_MATCHOPKEYWORD(MATCH_MGT); }

<CODE>Compatibility	{ ACTION_DMATYPEKEYWORD(DMATYPE_COMPAT); }
<CODE>TypeA		{ ACTION_DMATYPEKEYWORD(DMATYPE_TYPEA); }
<CODE>TypeB		{ ACTION_DMATYPEKEYWORD(DMATYPE_TYPEB); }
<CODE>TypeF		{ ACTION_DMATYPEKEYWORD(DMATYPE_TYPEF); }

<CODE>BusMaster		{ ACTION_BUSMASTERKEYWORD(1); }
<CODE>NotBusMaster	{ ACTION_BUSMASTERKEYWORD(0); }

<CODE>Transfer8		{ ACTION_XFERTYPEKEYWORD(XFERTYPE_XFER8); }
<CODE>Transfer16	{ ACTION_XFERTYPEKEYWORD(XFERTYPE_XFER16); }
<CODE>Transfer8_16	{ ACTION_XFERTYPEKEYWORD(XFERTYPE_XFER8_16); }

<CODE>ResourceConsumer	{ ACTION_RESOURCETYPEKEYWORD(RESOURCE_COMSUMER); }
<CODE>ResourceProducer	{ ACTION_RESOURCETYPEKEYWORD(RESOURCE_PRODUCER); }

<CODE>MinFixed		{ ACTION_MINKEYWORD(1); }
<CODE>MinNotFixed	{ ACTION_MINKEYWORD(0); }

<CODE>MaxFixed		{ ACTION_MAXKEYWORD(1); }
<CODE>MaxNotFixed	{ ACTION_MAXKEYWORD(0); }

<CODE>SubDecode		{ ACTION_DECODEKEYWORD(DECODE_SUBDECODE); }
<CODE>PosDecode		{ ACTION_DECODEKEYWORD(DECODE_POSDECODE); }

<CODE>ISAOnlyRanges	{ ACTION_RANGETYPEKEYWORD(RANGETYPE_ISAONLY); }
<CODE>NonISAOnlyRanges	{ ACTION_RANGETYPEKEYWORD(RANGETYPE_NONISAONLY); }
<CODE>EntireRange	{ ACTION_RANGETYPEKEYWORD(RANGETYPE_ENTIRE); }

<CODE>Cacheable		{ ACTION_MEMTYPEKEYWORD(MEMTYPE_CACHEABLE); }
<CODE>WriteCombining	{ ACTION_MEMTYPEKEYWORD(MEMTYPE_WRITECOMBINING); }
<CODE>Prefetchable	{ ACTION_MEMTYPEKEYWORD(MEMTYPE_PREFETCHABLE); }
<CODE>NonCacheable	{ ACTION_MEMTYPEKEYWORD(MEMTYPE_NONCACHEABLE); }

<CODE>ReadWrite		{ ACTION_READWRITEKEYWORD(READWRITE_READWRITE); }
<CODE>ReadOnly		{ ACTION_READWRITEKEYWORD(READWRITE_READONLY); }

<CODE>Edge		{ ACTION_INTERRUPTTYPEKEYWORD(INTERRUPTTYPE_EDGE); }
<CODE>Level		{ ACTION_INTERRUPTTYPEKEYWORD(INTERRUPTTYPE_LEVEL); }

<CODE>ActiveHigh	{ ACTION_INTERRUPTLEVELKEYWORD(INTERRUPTLEVEL_ACTIVEHIGH); }
<CODE>ActiveLow		{ ACTION_INTERRUPTLEVELKEYWORD(INTERRUPTLEVEL_ACTIVELOW); }

<CODE>Shared		{ ACTION_SHARETYPEKEYWORD(SHARETYPE_SHARED); }
<CODE>Exclusive		{ ACTION_SHARETYPEKEYWORD(SHARETYPE_EXCLUSIVE); }

<CODE>Decode16		{ ACTION_IODECODEKEYWORD(IODECODE_DECODE16); }
<CODE>Decode10		{ ACTION_IODECODEKEYWORD(IODECODE_DECODE10); }

<CODE>TypeTranslation	{ ACTION_TYPEKEYWORD(TYPE_TRANSLATION); }
<CODE>TypeStatic	{ ACTION_TYPEKEYWORD(TYPE_STATIC); }

<CODE>SparseTranslation	{ ACTION_TRANSLATIONKEYWORD(TRANSLATION_SPARSE); }
<CODE>DenseTranslation	{ ACTION_TRANSLATIONKEYWORD(TRANSLATION_DENSE); }

<CODE>AddressRangeMemory	{ ACTION_ADDRESSKEYWORD(ADDRESS_MEMORY); }
<CODE>AddressRangeReserved	{ ACTION_ADDRESSKEYWORD(ADDRESS_RESERVED); }
<CODE>AddressRangeNVS		{ ACTION_ADDRESSKEYWORD(ADDRESS_NVS); }
<CODE>AddressRangeACPI		{ ACTION_ADDRESSKEYWORD(ADDRESS_ACPI); }

 /* SuperName */
<CODE>Arg[0-6]		{ ACTION_VARIABLE(3, TOK_ARGTERM); }
<CODE>Local[0-7]	{ ACTION_VARIABLE(5, TOK_LOCALTERM); }
<CODE>Debug		{ return(TOK_DEBUGTERM); }

 /* Integer */
<CODE>0[xX]([0-9a-fA-F]{1,2})	{ ACTION_INTEGER(16, 1); }
<CODE>0[xX]([0-9a-fA-F]{3,4})	{ ACTION_INTEGER(16, 2); }
<CODE>0[xX]([0-9a-fA-F]{5,8})	{ ACTION_INTEGER(16, 4); }
<CODE>0[xX]([0-9a-fA-F]{9,16})	{ ACTION_INTEGER(16, 8); }
<CODE>[1-9][0-9]*	{ ACTION_INTEGER(10, 0); }
<CODE>0[0-7]*		{ ACTION_INTEGER(8, 0); }

 /* String */
<CODE>\"[^"]*\"		{
				size_t	i, len;

				len = yyleng - 2;
				yylval.lit.type = LITERAL_STRING;
				yylval.lit.str.ptr = (char *)malloc(len + 1);
				if (yylval.lit.str.ptr == NULL) {
					parser_fatal("No more room for String");
				}
				strncpy(yylval.lit.str.ptr, &yytext[1], len);
				yylval.lit.str.ptr[len] = '\0';
				for (i = 0; i < len; i++) {
					if (yylval.lit.str.ptr[i] == '\n') {
						parser_lineno++;
					}
				}
				return(TOK_ASCIICHARLIST);
			}

 /* Constant */
<CODE>Zero		{ ACTION_CONSTANT(0x00, 1, TOK_ZERO); }
<CODE>One		{ ACTION_CONSTANT(0x01, 1, TOK_ONE); }
<CODE>Ones		{ ACTION_CONSTANT(0xff, 1, TOK_ONES); }
<CODE>Revision		{ ACTION_CONSTANT(0x5b30, 2, TOK_REVISION); }

 /* DataObject */
<CODE>Buffer		{ return(TOK_BUFFER); }
<CODE>Package		{ return(TOK_PACKAGE); }

<CODE>EISAID		{ return(TOK_EISAID); }
<CODE>ResourceTemplate	{ return(TOK_RESOURCETEMPLATE); }
<CODE>Unicode		{ return(TOK_UNICODE); }

<CODE>DMA		{ return(TOK_DMA); }
<CODE>DWordIO		{ return(TOK_DWORDIO); }
<CODE>DWordMemory	{ return(TOK_DWORDMEMORY); }
<CODE>EndDependentFn	{ return(TOK_ENDDEPENDENTFN); }
<CODE>FixedIO		{ return(TOK_FIXEDIO); }
<CODE>Interrupt		{ return(TOK_INTERRUPT); }
<CODE>IO		{ return(TOK_IO); }
<CODE>IRQNoFlags	{ return(TOK_IRQNOFLAGS); }
<CODE>IRQ		{ return(TOK_IRQ); }
<CODE>Memory24		{ return(TOK_MEMORY24); }
<CODE>Memory32Fixed	{ return(TOK_MEMORY32FIXED); }
<CODE>Memory32		{ return(TOK_MEMORY32); }
<CODE>QWordIO		{ return(TOK_QWORDIO); }
<CODE>QWordMemory	{ return(TOK_QWORDMEMORY); }
<CODE>Register		{ return(TOK_REGISTER); }
<CODE>StartDependentFn	{ return(TOK_STARTDEPENDENTFN); }
<CODE>StartDependentFnNoPri	{ return(TOK_STARTDEPENDENTFNNOPRI); }
<CODE>VendorLong	{ return(TOK_VENDORLONG); }
<CODE>VendorShort	{ return(TOK_VENDORSHORT); }
<CODE>WordBusNumber	{ return(TOK_WORDBUSNUMBER); }
<CODE>WordIO		{ return(TOK_WORDIO); }

 /* ASL Names */
<CODE>[\\.^]		{ return(yytext[0]); }
<CODE>[A-Za-z_]([A-Za-z_0-9]{0,3})	{
				yylval.str = strdup(yytext);
				return(TOK_NAMESEG);
			}
 /* Error */
<CODE>[A-Z_][A-Z_0-9]{4,}	{
				parser_error("<%s> is too long", yytext);
				return(TOK_ERROR);
}
<CODE>[^/*\n \t,{}()0-9a-fA-Z".^_]+	{
				parser_error("<%s> is unexpected", yytext);
				return(TOK_ERROR);
			}
%%
