%{
/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>

#include <err.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "aslc.h"
#include "aslcsymtree.h"

#ifdef DEBUG
#define YYDEBUG 1
#define PARSER_DEBUGPRINT(phase, x...)	do {				\
	if (parser_phase <= phase && parser_debuglevel) {		\
		printf(x);						\
	}								\
} while (0)

#else
#define PARSER_DEBUGPRINT(phase, x...)
#endif

#define PARSER_SYMTAB_NEW(dest, name, type)	do {			\
	dest = symtab_find((const char *)name);				\
	if (dest != NULL) {						\
		parser_warn("<%s> is already defined", name);		\
	} else {							\
		dest = symtab_new((const char *)name, type);		\
	}								\
} while (0)

#define PARSER_SYMTAB_FIND(dest, name, typestr)	do {			\
	dest = symtab_find((const char *)name);				\
	if (dest == NULL) {						\
		snprintf(parser_msgbuf, PARSER_MSGBUFSIZE,		\
		    "%s<%s> is not defined", typestr, name);		\
		yyerror(parser_msgbuf);					\
	}								\
} while (0)

int		 parser_phase = 0;
int		 parser_debuglevel = 0;
int		 parser_not_initialized = 1;

#define PARSER_MSGBUFSIZE	256
static char	 parser_msgbuf[PARSER_MSGBUFSIZE];

static char	*aml_filename = NULL;

static char	*parser_strcat(const char *str1, const char *str2);

static void	 parser_scope_enter(const char *name);
static void	 parser_scope_exit(void);
static char	*parser_scope_fullpath(const char *name);

%}

%union {
	char			*str;
	u_int32_t		 num;	/* XXX should be 64 bits for QWord */
	union asl_literal	 lit;
	struct asl_symtab	*sym;
}

/*
 * Token
 */

%token		TOK_DEFINITIONBLOCK

%token		TOK_INCLUDE
%token		TOK_EXTERNAL

%token		TOK_BANKFIELD
%token		TOK_OFFSET
%token		TOK_ACCESSAS
%token		TOK_CREATEBITFIELD
%token		TOK_CREATEBYTEFIELD
%token		TOK_CREATEDWORDFIELD
%token		TOK_CREATEFIELD
%token		TOK_CREATEQWORDFIELD
%token		TOK_CREATEWORDFIELD
%token		TOK_DATATABLEREGION
%token		TOK_DEVICE
%token		TOK_EVENT
%token		TOK_FIELD
%token		TOK_INDEXFIELD
%token		TOK_METHOD
%token		TOK_MUTEX
%token		TOK_OPERATIONREGION
%token		TOK_POWERRESOURCE
%token		TOK_PROCESSOR
%token		TOK_THERMALZONE
%token		TOK_ALIAS
%token		TOK_NAME
%token		TOK_SCOPE

%token		TOK_BREAK
%token		TOK_BREAKPOINT
%token		TOK_CONTINUE
%token		TOK_FATAL
%token		TOK_IF
%token		TOK_ELSE
%token		TOK_ELSEIF
%token		TOK_LOAD
%token		TOK_NOOP
%token		TOK_NOTIFY
%token		TOK_RELEASE
%token		TOK_RESET
%token		TOK_RETURN
%token		TOK_SIGNAL
%token		TOK_SLEEP
%token		TOK_STALL
%token		TOK_SWITCH
%token		TOK_CASE
%token		TOK_DEFAULT
%token		TOK_UNLOAD
%token		TOK_WHILE

%token		TOK_ACQUIRE
%token		TOK_ADD
%token		TOK_AND
%token		TOK_BUFF
%token		TOK_CONCATENATE
%token		TOK_CONCATENATERESTEMPLATE
%token		TOK_CONDREFOF
%token		TOK_COPY
%token		TOK_DECREMENT
%token		TOK_DECSTR
%token		TOK_DEREFOF
%token		TOK_DIVIDE
%token		TOK_FINDSETLEFTBIT
%token		TOK_FINDSETRIGHTBIT
%token		TOK_FROMBCD
%token		TOK_HEXSTR
%token		TOK_INCREMENT
%token		TOK_INDEX
%token		TOK_INT
%token		TOK_LAND
%token		TOK_LEQUAL
%token		TOK_LGREATER
%token		TOK_LGREATEREQUAL
%token		TOK_LLESS
%token		TOK_LLESSEQUAL
%token		TOK_LNOT
%token		TOK_LNOTEQUAL
%token		TOK_LOADTABLE
%token		TOK_LOR
%token		TOK_MATCH
%token		TOK_MID
%token		TOK_MOD
%token		TOK_MULTIPLY
%token		TOK_NAND
%token		TOK_NOR
%token		TOK_NOT
%token		TOK_OBJECTTYPE
%token		TOK_OR
%token		TOK_REFOF
%token		TOK_SHIFTLEFT
%token		TOK_SHIFTRIGHT
%token		TOK_SIZEOF
%token		TOK_STORE
%token		TOK_STRING
%token		TOK_SUBTRACT
%token		TOK_TOBCD
%token		TOK_WAIT
%token		TOK_XOR

%token <num>	TOK_OBJECTTYPEKEYWORD
%token <num>	TOK_ACCESSTYPEKEYWORD
%token <num>	TOK_ACCESSATTRIBKEYWORD
%token <num>	TOK_LOCKRULEKEYWORD
%token <num>	TOK_UPDATERULEKEYWORD
%token <num>	TOK_REGIONSPACEKEYWORD
%token		TOK_FFIXEDHW
%token <num>	TOK_SERIALIZERULEKEYWORD
%token <num>	TOK_MATCHOPKEYWORD
%token <num>	TOK_DMATYPEKEYWORD
%token <num>	TOK_BUSMASTERKEYWORD
%token <num>	TOK_XFERTYPEKEYWORD
%token <num>	TOK_RESOURCETYPEKEYWORD
%token <num>	TOK_MINKEYWORD
%token <num>	TOK_MAXKEYWORD
%token <num>	TOK_DECODEKEYWORD
%token <num>	TOK_RANGETYPEKEYWORD
%token <num>	TOK_MEMTYPEKEYWORD
%token <num>	TOK_READWRITEKEYWORD
%token <num>	TOK_INTERRUPTTYPEKEYWORD
%token <num>	TOK_INTERRUPTLEVELKEYWORD
%token <num>	TOK_SHARETYPEKEYWORD
%token <num>	TOK_IODECODEKEYWORD
%token <num>	TOK_TYPEKEYWORD
%token <num>	TOK_TRANSLATIONKEYWORD
%token <num>	TOK_ADDRESSKEYWORD

%token <num>	TOK_ARGTERM TOK_LOCALTERM
%token		TOK_DEBUGTERM

%token <lit>	TOK_ZERO TOK_ONE TOK_ONES TOK_REVISION
%token <lit>	TOK_BYTECONST TOK_WORDCONST TOK_DWORDCONST TOK_QWORDCONST

%token <lit>	TOK_ASCIICHARLIST

%token		TOK_BUFFER
%token		TOK_PACKAGE

%token		TOK_EISAID
%token		TOK_RESOURCETEMPLATE
%token		TOK_UNICODE
%token		TOK_DMA
%token		TOK_DWORDIO
%token		TOK_DWORDMEMORY
%token		TOK_ENDDEPENDENTFN
%token		TOK_FIXEDIO
%token		TOK_INTERRUPT
%token		TOK_IO
%token		TOK_IRQNOFLAGS
%token		TOK_IRQ
%token		TOK_MEMORY24
%token		TOK_MEMORY32FIXED
%token		TOK_MEMORY32
%token		TOK_QWORDIO
%token		TOK_QWORDMEMORY
%token		TOK_REGISTER
%token		TOK_STARTDEPENDENTFN
%token		TOK_STARTDEPENDENTFNNOPRI
%token		TOK_VENDORLONG
%token		TOK_VENDORSHORT
%token		TOK_WORDBUSNUMBER
%token		TOK_WORDIO

%token <str>	TOK_NAMESEG

%token		TOK_ERROR

/*
 * Type
 */

%type <num>	IntegerData
%type <str>	StringData
%type <sym>	UserTerm

%type <sym>	CreateByteFieldTerm
%type <sym>	MethodTerm
%type <num>	MethodNumArgs
%type <num>	MethodSerializeRule
%type <num>	MethodSyncLevel

%type <num>	ObjectTypeKeyword
%type <num>	AccessTypeKeyword
%type <num>	AccessAttribKeyword
%type <num>	LockRuleKeyword
%type <num>	UpdateRuleKeyword
%type <num>	RegionSpaceKeyword
%type <num>	AddressSpaceKeyword
%type <num>	UserDefRegionSpace
%type <num>	SerializeRuleKeyword
%type <num>	MatchOpKeyword
%type <num>	DMATypeKeyword
%type <num>	BusMasterKeyword
%type <num>	XferTypeKeyword
%type <num>	ResourceTypeKeyword
%type <num>	MinKeyword
%type <num>	MaxKeyword
%type <num>	DecodeKeyword
%type <num>	RangeTypeKeyword
%type <num>	MemTypeKeyword
%type <num>	ReadWriteKeyword
%type <num>	InterruptTypeKeyword
%type <num>	InterruptLevelKeyword
%type <num>	ShareTypeKeyword
%type <num>	IODecodeKeyword
%type <num>	TypeKeyword
%type <num>	TranslationKeyword
%type <num>	AddressKeyword

%type <lit>	Integer
%type <lit>	ByteConst
%type <lit>	WordConst
%type <lit>	DWordConst
%type <lit>	QWordConst

%type <lit>	String
%type <lit>	ConstTerm

%type <lit>	ByteConstExpr
%type <lit>	WordConstExpr
%type <lit>	DWordConstExpr
%type <lit>	QWordConstExpr
%type <lit>	ConstExprTerm

%type <str>	NameSeg
%type <str>	NameString
%type <str>	PrefixPath
%type <str>	NamePath
%type <str>	NonEmptyNamePath
%type <str>	NamePathTail

%%

ASLCode
	:	{
			symtree_moveptr(symtree_new(SYMTYPE_DefinitionBlock,
						    "DefinitionBlockTerm"));
		}
	  DefinitionBlockTerm
		{
			symtree_upptr();
		}
	;


DefinitionBlockTerm
	: TOK_DEFINITIONBLOCK
	  '(' String ',' String ',' ByteConst ',' String ',' String ',' DWordConst ')'
		{
			struct asl_symtree	*ptr;

			if (strlen($5.str.ptr) > 4) {
				parser_error("TableSignature<%s> is too long",
					     $5.str.ptr);
			}
			if (strlen($9.str.ptr) > 6) {
				parser_error("OEMID<%s> is too long",
					     $9.str.ptr);
			}
			if (strlen($11.str.ptr) > 8) {
				parser_error("TableID<%s> is too long",
					     $11.str.ptr);
			}
			aml_filename = strdup($3.str.ptr);
			ptr = symtree_getptr();
			ptr->symbol.DefinitionBlock.AMLFileName = $3.str.ptr;
			ptr->symbol.DefinitionBlock.TableSignature = $5.str.ptr;
			ptr->symbol.DefinitionBlock.ComplianceRevision = $7.num.val;
			ptr->symbol.DefinitionBlock.OEMID = $9.str.ptr;
			ptr->symbol.DefinitionBlock.TableID = $11.str.ptr;
			ptr->symbol.DefinitionBlock.OEMRevision = $13.num.val;
			symtree_purge(NULL);

			symtree_upptr();
			ptr = symtree_new(SYMTYPE_TermList, "TermList");
			symtree_moveptr(ptr);
		}
	  '{' TermList '}'
	;


TermList
	: Nothing | Term TermList
	;
Term
	: Object | Type1Opcode | Type2Opcode
	;


CompilerDirective
	: IncludeTerm | ExternalTerm
	;


ObjectList
	: Nothing | Object ObjectList
	;
Object
	: CompilerDirective | NamedObject | NameSpaceModifier
	;

DataObject
	: BufferData | PackageData | IntegerData | StringData
	;
DataRefObject
	: DataObject /* XXX | ObjectReference | DDBHandle */
	| NameString /* XXX Named object by user (not in Spec.) */
	;


/* XXX
ComputationalData
	: BufferData | IntegerData | StringData
	;
*/
BufferData
	: Type5Opcode | BufferTerm
	;
PackageData
	: PackageTerm
	;
IntegerData
	: Type3Opcode			{ $$ = 0; }
	| Integer			{ $$ = $1.num.val; }
	| ConstTerm			{ $$ = $1.num.val; }
	;
StringData
	: Type4Opcode			{ $$ == NULL; }
	| String			{ $$ = $1.str.ptr; }
	;


NamedObject
	: BankFieldTerm | CreateBitFieldTerm | CreateByteFieldTerm
	| CreateDWordFieldTerm | CreateFieldTerm
	| CreateQWordFieldTerm | CreateWordFieldTerm
	| DataRegionTerm | DeviceTerm | EventTerm | FieldTerm
	| IndexFieldTerm | MethodTerm | MutexTerm | OpRegionTerm
	| PowerResTerm | ProcessorTerm | ThermalZoneTerm
	;


NameSpaceModifier
	: AliasTerm | NameTerm | ScopeTerm
	;


UserTerm
	: NameString '('
		{
			symtree_moveptr(symtree_new(SYMTYPE_UserTerm,
						    "UserTerm"));
		}
	  ArgList ')'
		{
			symtree_upptr();
		}
	;
ArgList
	: Nothing
	| TermArg ArgListTail
	;
ArgListTail
	: Nothing
	| ',' TermArg ArgListTail
	;
TermArg
	: Type2Opcode | DataRefObject | ArgTerm | LocalTerm
	;
Target
	: Nothing | SuperName
	;


Type1Opcode
	: BreakTerm | BreakPointTerm | ContinueTerm | FatalTerm
	| IfElseTerm | LoadTerm | NoOpTerm | NotifyTerm
	| ReleaseTerm | ResetTerm | ReturnTerm | SignalTerm
	| SleepTerm | StallTerm | SwitchTerm | UnloadTerm
	| WhileTerm
	/*
	 * A Type1OpCode term can only be used standing alone on a
	 * line of ASL code; because these types of terms do not
	 * return a value so they cannot be used as a term in an
	 * expression.
	 */
	;


Type2Opcode
	: AcquireTerm | AddTerm | AndTerm | BuffTerm | ConcatTerm
	| ConcatResTerm | CondRefOfTerm | CopyTerm | DecTerm
	| DecStrTerm | DerefOfTerm | DivideTerm
	| FindSetLeftBitTerm | FindSetRightBitTerm | FromBCDTerm
	| HexStrTerm | IncTerm | IndexTerm | IntTerm | LAndTerm
	| LEqualTerm | LGreaterTerm | LGreaterEqualTerm
	| LLessTerm | LLessEqualTerm | LNotTerm | LNotEqualTerm
	| LoadTableTerm | LOrTerm | MatchTerm | MidTerm | ModTerm
	| MultiplyTerm | NAndTerm | NOrTerm | NotTerm
	| ObjectTypeTerm | OrTerm | RefOfTerm | ShiftLeftTerm
	| ShiftRightTerm | SizeOfTerm | StoreTerm | StringTerm
	| SubtractTerm | ToBCDTerm | WaitTerm | XorTerm | UserTerm
	/*
	 * A Type2Opcode term returns a value that can be used
	 * inan expression.
	 */
	;


Type3Opcode
	: AddTerm | AndTerm | DecTerm | DivideTerm | EISAIDTerm
	| FindSetLeftBitTerm | FindSetRightBitTerm | FromBCDTerm
	| IncTerm | IndexTerm | IntTerm | LAndTerm | LEqualTerm
	| LGreaterTerm | LGreaterEqualTerm | LLessTerm
	| LLessEqualTerm | LNotTerm | LNotEqualTerm | LOrTerm
	| MatchTerm | ModTerm | MultiplyTerm | NAndTerm | NOrTerm
	| NotTerm | OrTerm | ShiftLeftTerm | ShiftRightTerm
	| SubtractTerm | ToBCDTerm | XorTerm
	/*
	 * A Type3Opcode evaluates to an Integer, can't have a
	 * destination and must have either Type3Opcode,
	 * Type4Opcode, Type5Opcode, ConstExprTerm, Integer,
	 * BufferTerm, Package or String for all arguments.
	 */
	;


Type4Opcode
	: ConcatTerm | DecStrTerm | HexStrTerm | MidTerm | StringTerm
	/*
	 * A Type4Opcode evaluates to an String, can't have a
	 * destination and must have either Type3Opcode,
	 * Type4Opcode, Type5Opcode, ConstExprTerm, Integer,
	 * BufferTerm, PackageTerm or String for all arguments.
	 */
	;


Type5Opcode
	: BuffTerm | ConcatTerm | ConcatResTerm | MidTerm
	| ResourceTemplateTerm | UnicodeTerm
	/*
	 * A Type5Opcode evaluates to a BufferTerm, can't
	 * have a destination and must have either Type3Opcode,
	 * Type4Opcode, Type5Opcode, ConstExprTerm, Integer,
	 * BufferTerm, PackageTerm or String for all arguments.
	 */
	;


Type6Opcode
	: RefOfTerm | DerefOfTerm | IndexTerm | UserTerm
	;


IncludeTerm
	: TOK_INCLUDE '(' StringData ')'
	;


ExternalTerm
	: TOK_EXTERNAL '(' NameString ',' ObjType ')'
	;
ObjType
	: Nothing | ObjectTypeKeyword
	;


/*
 * Named Object
 */

BankFieldTerm
	: TOK_BANKFIELD
	  '(' NameString ',' NameString ',' TermArg ',' AccessTypeKeyword ',' LockRuleKeyword ',' UpdateRuleKeyword ')'
	  '{' FieldUnitList '}'
	;


FieldUnitList
	: Nothing | FieldUnit FieldUnitListTail
	;
FieldUnitListTail
	: Nothing | ',' FieldUnit FieldUnitListTail
	;


FieldUnit
	: FieldUnitEntry | OffsetTerm | AccessAsTerm
	;
FieldUnitEntry
	: FieldUnitEntryName ',' Integer
	;
FieldUnitEntryName
	: Nothing | NameSeg
	;


OffsetTerm
	: TOK_OFFSET
	  '(' IntegerData ')'
	;


AccessAsTerm
	: TOK_ACCESSAS
	  '(' AccessTypeKeyword AccessAttribute ')'
	;
AccessAttribute
	: Nothing | ',' ByteConstExpr | ',' AccessAttribKeyword
	;


CreateBitFieldTerm
	: TOK_CREATEBITFIELD
	  '(' TermArg ',' TermArg ',' NameString ')'
	;


CreateByteFieldTerm
	: TOK_CREATEBYTEFIELD
		{
			symtree_moveptr(symtree_new(SYMTYPE_CreateByteFieldTerm,
						    "CreateByteFieldTerm"));
		}
	  '(' TermArg ',' TermArg ',' NameString ')'
		{
			PARSER_SYMTAB_NEW($$, parser_scope_fullpath($8), SYM_FIELD);
			symtree_upptr();

		}
	;


CreateDWordFieldTerm
	: TOK_CREATEDWORDFIELD
	  '(' TermArg ',' TermArg ',' NameString ')'
	;


CreateFieldTerm
	: TOK_CREATEFIELD
	  '(' TermArg ',' TermArg ',' TermArg ',' NameString ')'
	;


CreateQWordFieldTerm
	: TOK_CREATEQWORDFIELD
	  '(' TermArg ',' TermArg ',' NameString ')'
	;


CreateWordFieldTerm
	: TOK_CREATEWORDFIELD
	  '(' TermArg ',' TermArg ',' NameString ')'
	;


DataRegionTerm
	: TOK_DATATABLEREGION
	  '(' NameString ',' TermArg ',' TermArg ',' TermArg ')'
	;


DeviceTerm
	: TOK_DEVICE
	  '(' NameString ')'
	  '{' ObjectList '}'
	;


EventTerm
	: TOK_EVENT
	  '(' NameString ')'
	;


FieldTerm
	: TOK_FIELD
	  '(' NameString ',' AccessTypeKeyword ',' LockRuleKeyword ',' UpdateRuleKeyword ')'
	  '{' FieldUnitList '}'
	;


IndexFieldTerm
	: TOK_INDEXFIELD
	  '(' NameString ',' NameString ',' AccessTypeKeyword ',' LockRuleKeyword ',' UpdateRuleKeyword ')'
	  '{' FieldUnitList '}'
	;


MethodTerm
	: TOK_METHOD
		{
			symtree_moveptr(symtree_new(SYMTYPE_MethodTerm,
						    "MethodTerm"));
		}
	  '(' NameString MethodNumArgs MethodSerializeRule MethodSyncLevel ')'
		{
			struct asl_symtree	*ptr;

			PARSER_SYMTAB_NEW($$, parser_scope_fullpath($4), SYM_METHOD);
			symtab_method_setflags($$, (int)($5 | $6 << 3));

			parser_scope_enter($4);

			ptr = symtree_getptr();
			ptr->symbol.Method.MethodName = $4;
			ptr->symbol.Method.NumArgs = $5;
			ptr->symbol.Method.SerializeRule = $6;
			ptr->symbol.Method.SyncLevel = $7;
			symtree_purge(NULL);

			symtree_moveptr(symtree_new(SYMTYPE_TermList,
						    "TermList"));
		}
	  '{' TermList '}'
		{
			parser_scope_exit();

			symtree_upptr();
			symtree_upptr();
		}
	;
MethodNumArgs
	: Nothing			{ $$ = 0; }
	| ',' ByteConstExpr		{ $$ = $2.num.val; }
	;
MethodSerializeRule
	: Nothing			{ $$ = 0; }
	| ',' SerializeRuleKeyword	{ $$ = $2; }
	;
MethodSyncLevel
	: Nothing			{ $$ = 0; }
	| ',' ByteConstExpr		{ $$ = $2.num.val; }
	;


MutexTerm
	: TOK_MUTEX
	  '(' NameString ',' ByteConstExpr ')'
	;


OpRegionTerm
	: TOK_OPERATIONREGION
	  '(' NameString ',' RegionSpaceKeyword ',' TermArg ',' TermArg ')'
	;


PowerResTerm
	: TOK_POWERRESOURCE
	  '(' NameString ',' ByteConstExpr ',' WordConstExpr ')'
	  '{' ObjectList '}'
	;


ProcessorTerm
	: TOK_PROCESSOR
	  '(' NameString ',' ByteConstExpr ',' PBlockAddress ',' PBlockLength ')'
	  '{' ObjectList '}'
	;
PBlockAddress
	: DWordConstExpr | Nothing
	;
PBlockLength
	: ByteConstExpr | Nothing
	;


ThermalZoneTerm
	: TOK_THERMALZONE
	  '(' NameString ')'
	  '{' ObjectList '}'
	;


AliasTerm
	: TOK_ALIAS
	  '(' NameString ',' NameString ')'
	;


NameTerm
	: TOK_NAME
	  '(' NameString ',' DataRefObject ')'
	;


ScopeTerm
	: TOK_SCOPE
	  '(' NameString ')'
	  '{' ObjectList '}'
	;


/*
 * Type1Opcode
 */

BreakTerm
	: TOK_BREAK
	;


BreakPointTerm
	: TOK_BREAKPOINT
	;


ContinueTerm
	: TOK_CONTINUE
	;


FatalTerm
	: TOK_FATAL
	  '(' ByteConstExpr ',' DWordConstExpr ',' TermArg ')'
	;


IfElseTerm
	: IfTerm ElseTerm
	;


IfTerm
	: TOK_IF
	  '(' TermArg ')'
	  '{' TermList '}'
	;


ElseTerm
	: Nothing
	| TOK_ELSE
	  '{' TermList '}'
	| TOK_ELSEIF
	  '{' TermList '}'
	  ElseTerm
	;


LoadTerm
	: TOK_LOAD
	  '(' NameString ',' SuperName ')'
	;


NoOpTerm
	: TOK_NOOP
	;


NotifyTerm
	: TOK_NOTIFY
	  '(' SuperName ',' TermArg ')'
	;


ReleaseTerm
	: TOK_RELEASE
	  '(' SuperName ')'
	;


ResetTerm
	: TOK_RESET
	  '(' SuperName ')'
	;


ReturnTerm
	: TOK_RETURN
	  '(' TermArg ')'
	;


SignalTerm
	: TOK_SIGNAL
	  '(' SuperName ')'
	;


SleepTerm
	: TOK_SLEEP
	  '(' TermArg ')'
	;


StallTerm
	: TOK_STALL
	  '(' TermArg ')'
	;


SwitchTerm
	: TOK_SWITCH
	  '(' TermArg ')'
	  '{' CaseTermList '}'
	;


CaseTermList
	: Nothing | CaseTerm | DefaultTerm DefaultTermList | CaseTerm CaseTermList
	;


DefaultTermList
	: Nothing | CaseTerm | CaseTerm DefaultTermList
	;


CaseTerm
	: TOK_CASE
	  '(' DataObject ')'
	  '{' TermList '}'
	;


DefaultTerm
	: TOK_DEFAULT
	  '{' TermList '}'
	;


UnloadTerm
	: TOK_UNLOAD
	  '(' SuperName ')'
	;


WhileTerm
	: TOK_WHILE
		{
			symtree_moveptr(symtree_new(SYMTYPE_WhileTerm,
						    "WhileTerm"));
		}
	  '(' TermArg ')' 
		{
			symtree_moveptr(symtree_new(SYMTYPE_TermList,
						    "TermList"));
		}
	  '{' TermList '}'
		{
			symtree_upptr();
			symtree_upptr();
		}
	;


/*
 * Type2Opcode
 */

AcquireTerm
	: TOK_ACQUIRE
	  '(' SuperName ',' WordConstExpr ')'
	;


AddTerm
	: TOK_ADD
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;
OptionalTarget
	: Nothing
	| ',' Target


AndTerm
	: TOK_AND
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


BuffTerm
	: TOK_BUFF
	  '(' TermArg OptionalTarget ')'
	;


ConcatTerm
	: TOK_CONCATENATE
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


ConcatResTerm
	: TOK_CONCATENATERESTEMPLATE
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


CondRefOfTerm
	: TOK_CONDREFOF
	  '(' SuperName OptionalTarget ')'
	;


CopyTerm
	: TOK_COPY
	  '(' TermArg ',' CopyTermResult ')'
	;
CopyTermResult
	: NameString | LocalTerm | ArgTerm
	;


DecTerm
	: TOK_DECREMENT
		{
			symtree_moveptr(symtree_new(SYMTYPE_DecTerm, "DecTerm"));
		}
	  '(' SuperName ')'
		{
			symtree_upptr();
		}
	;


DecStrTerm
	: TOK_DECSTR
	  '(' TermArg OptionalTarget ')'
	;


DerefOfTerm
	: TOK_DEREFOF
	  '(' TermArg ')'
	;


DivideTerm
	: TOK_DIVIDE
	  '(' TermArg ',' TermArg DivideTermRemainder  DivideTermResult ')'
	;
DivideTermRemainder
	: OptionalTarget
	;
DivideTermResult
	: OptionalTarget
	;

FindSetLeftBitTerm
	: TOK_FINDSETLEFTBIT
	  '(' TermArg OptionalTarget ')'
	;


FindSetRightBitTerm
	: TOK_FINDSETRIGHTBIT
	  '(' TermArg OptionalTarget ')'
	;


FromBCDTerm
	: TOK_FROMBCD
	  '(' TermArg OptionalTarget ')'
	;


HexStrTerm
	: TOK_HEXSTR
	  '(' TermArg OptionalTarget ')'
	;


IncTerm
	: TOK_INCREMENT
	  '(' SuperName ')'
	;


IndexTerm
	: TOK_INDEX
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


IntTerm
	: TOK_INT
	  '(' TermArg OptionalTarget ')'
	;


LAndTerm
	: TOK_LAND
	  '(' TermArg ',' TermArg ')'
	;


LEqualTerm
	: TOK_LEQUAL
	  '(' TermArg ',' TermArg ')'
	;


LGreaterTerm
	: TOK_LGREATER
	  '(' TermArg ',' TermArg ')'
	;


LGreaterEqualTerm
	: TOK_LGREATEREQUAL
	  '(' TermArg ',' TermArg ')'
	;


LLessTerm
	: TOK_LLESS
	  '(' TermArg ',' TermArg ')'
	;


LLessEqualTerm
	: TOK_LLESSEQUAL
	  '(' TermArg ',' TermArg ')'
	;


LNotTerm
	: TOK_LNOT
	  '(' TermArg ')'
	;


LNotEqualTerm
	: TOK_LNOTEQUAL
	  '(' TermArg ',' TermArg ')'
	;


LoadTableTerm
	: TOK_LOADTABLE
	  '(' TermArg ',' TermArg ',' TermArg OptionalTermArg OptionalTermArg OptionalTermArg ')'
	;
OptionalTermArg
	: Nothing | TermArg
	;


LOrTerm
	: TOK_LOR
	  '(' TermArg ',' TermArg ')'
	;


MatchTerm
	: TOK_MATCH
	  '(' TermArg ',' MatchOpKeyword ',' TermArg ',' MatchOpKeyword ',' TermArg ',' TermArg ')'
	;


MidTerm
	: TOK_MID
	  '(' TermArg ',' TermArg ',' TermArg OptionalTarget ')'
	;


ModTerm
	: TOK_MOD
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


MultiplyTerm
	: TOK_MULTIPLY
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;



NAndTerm
	: TOK_NAND
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


NOrTerm
	: TOK_NOR
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


NotTerm
	: TOK_NOT
	  '(' TermArg OptionalTarget ')'
	;


ObjectTypeTerm
	: TOK_OBJECTTYPE
	  '(' SuperName ')'
	;


OrTerm
	: TOK_OR
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


RefOfTerm
	: TOK_REFOF
		{
			symtree_moveptr(symtree_new(SYMTYPE_RefOfTerm,
						    "RefOfTerm"));
		}
	  '(' SuperName ')'
		{
			symtree_upptr();
		}
	;


ShiftLeftTerm
	: TOK_SHIFTLEFT
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


ShiftRightTerm
	: TOK_SHIFTRIGHT
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


SizeOfTerm
	: TOK_SIZEOF
	  '(' SuperName ')'
	;


StoreTerm
	: TOK_STORE
		{
			symtree_moveptr(symtree_new(SYMTYPE_StoreTerm,
						    "StoreTerm"));
		}
	  '(' TermArg ',' SuperName ')'
		{
			symtree_upptr();
		}
	;


StringTerm
	: TOK_STRING
	  '(' TermArg StringTermLength OptionalTarget ')'
	;
StringTermLength
	: Nothing | ',' TermArg
	;


SubtractTerm
	: TOK_SUBTRACT
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


ToBCDTerm
	: TOK_TOBCD
	  '(' TermArg OptionalTarget ')'
	;


WaitTerm
	: TOK_WAIT
	  '(' SuperName ',' TermArg ')'
	;


XorTerm
	: TOK_XOR
	  '(' TermArg ',' TermArg OptionalTarget ')'
	;


/*
 * Keyword
 */

ObjectTypeKeyword
	: TOK_OBJECTTYPEKEYWORD
	;


AccessTypeKeyword
	: TOK_ACCESSTYPEKEYWORD
	;
AccessAttribKeyword
	: TOK_ACCESSATTRIBKEYWORD
	/* Note: AccessAttribKeywords are for SMBus BufferAcc only. */
	;


LockRuleKeyword
	: TOK_LOCKRULEKEYWORD
	;
UpdateRuleKeyword
	: TOK_UPDATERULEKEYWORD
	;


RegionSpaceKeyword
	: UserDefRegionSpace | TOK_REGIONSPACEKEYWORD
	;
AddressSpaceKeyword
	: RegionSpaceKeyword
	| TOK_FFIXEDHW			{ $$ = 0; /* XXX */ }
	;
UserDefRegionSpace
	: IntegerData
	;


SerializeRuleKeyword
	: TOK_SERIALIZERULEKEYWORD
	;


MatchOpKeyword
	: TOK_MATCHOPKEYWORD
	;


DMATypeKeyword
	: TOK_DMATYPEKEYWORD
	;
BusMasterKeyword
	: TOK_BUSMASTERKEYWORD
	;
XferTypeKeyword
	: TOK_XFERTYPEKEYWORD
	;


ResourceTypeKeyword
	: TOK_RESOURCETYPEKEYWORD
	;
MinKeyword
	: TOK_MINKEYWORD
	;
MaxKeyword
	: TOK_MAXKEYWORD
	;
DecodeKeyword
	: TOK_DECODEKEYWORD
	;
RangeTypeKeyword
	: TOK_RANGETYPEKEYWORD
	;
MemTypeKeyword
	: TOK_MEMTYPEKEYWORD
	;
ReadWriteKeyword
	: TOK_READWRITEKEYWORD
	;
InterruptTypeKeyword
	: TOK_INTERRUPTTYPEKEYWORD
	;
InterruptLevelKeyword
	: TOK_INTERRUPTLEVELKEYWORD
	;
ShareTypeKeyword
	: TOK_SHARETYPEKEYWORD
	;
IODecodeKeyword
	: TOK_IODECODEKEYWORD
	;
TypeKeyword
	: TOK_TYPEKEYWORD
	;
TranslationKeyword
	: TOK_TRANSLATIONKEYWORD
	;
AddressKeyword
	: TOK_ADDRESSKEYWORD
	;

/*
 * SuperName
 */

SuperName
	: NameString | ArgTerm | LocalTerm | DebugTerm | Type6Opcode | UserTerm
	;
ArgTerm
	: TOK_ARGTERM
		{
			struct asl_symtree	*ptr;

			ptr = symtree_new(SYMTYPE_ArgTerm, "ArgTerm");
			ptr->symbol.common.val = $1;
		}
	;
LocalTerm
	: TOK_LOCALTERM
		{
			struct asl_symtree	*ptr;

			ptr = symtree_new(SYMTYPE_LocalTerm, "LocalTerm");
			ptr->symbol.common.val = $1;
		}
	;
DebugTerm
	: TOK_DEBUGTERM
	;

/*
 * Integer
 */

Integer
	: ByteConst | WordConst | DWordConst | QWordConst
	;
ByteConst
	: TOK_BYTECONST
		{
			struct asl_symtree	*ptr;

			$$ = $1;
			ptr = symtree_new(SYMTYPE_ByteConst, "ByteConst");
			ptr->symbol.common.val = $1.num.val;
		}
	;
WordConst
	: TOK_WORDCONST
		{
			struct asl_symtree	*ptr;

			$$ = $1;
			ptr = symtree_new(SYMTYPE_WordConst, "WordConst");
			ptr->symbol.common.val = $1.num.val;
		}
	;
DWordConst
	: TOK_DWORDCONST
		{
			struct asl_symtree	*ptr;

			$$ = $1;
			ptr = symtree_new(SYMTYPE_DWordConst, "DWordConst");
			ptr->symbol.common.val = $1.num.val;
		}
	;
QWordConst
	: TOK_QWORDCONST
		{
			struct asl_symtree	*ptr;

			$$ = $1;
			parser_warn("QWord constant is not supported yet");

			ptr = symtree_new(SYMTYPE_QWordConst, "QWordConst");
			ptr->symbol.common.val = $1.num.val;
		}
	;


/*
 * String
 */

String
	: TOK_ASCIICHARLIST
		{
			struct asl_symtree	*ptr;

			$$ = $1;
			ptr = symtree_new(SYMTYPE_String, $$.str.ptr);
			ptr->symbol.String.ptr = $$.str.ptr;
		}
	;


ConstTerm
	: TOK_ZERO | TOK_ONE | TOK_ONES | TOK_REVISION
	;
/* XXX
Boolean
	: True | False
	;
True
	: TOK_ONES
	;
False
	: TOK_ZERO
	;
*/


ByteConstExpr
	: Type3Opcode { $$.num.val = 0; /* XXX */ } | ConstExprTerm | Integer
	;
WordConstExpr
	: Type3Opcode { $$.num.val = 0; /* XXX */ } | ConstExprTerm | Integer
	;
DWordConstExpr
	: Type3Opcode { $$.num.val = 0; /* XXX */ } | ConstExprTerm | Integer
	;
QWordConstExpr
	: Type3Opcode { $$.num.val = 0; /* XXX */ } | ConstExprTerm | Integer
	;
ConstExprTerm
	: TOK_ZERO | TOK_ONE | TOK_ONES
	;


/*
 * DataObject
 */

BufferTerm
	: TOK_BUFFER
		{
			symtree_moveptr(symtree_new(SYMTYPE_BufferTerm,
						    "BufferTerm"));
		}
	;
	  '(' BufferSize ')' 
		{
			symtree_moveptr(symtree_new(SYMTYPE_BufferBlock,
						     "BufferBlock"));
		}
	  '{' BufferBlock '}'
		{
			symtree_upptr();
			symtree_upptr();
		}
	;
BufferSize
	: Nothing | TermArg
	;
BufferBlock
	: StringData | ByteList
	;


ByteList
	: Nothing | ByteConstExpr ByteListTail
	;
ByteListTail
	: Nothing | ',' ByteConstExpr ByteListTail
	;


DWordList
	: Nothing | DWordConstExpr DWordListTail
	;
DWordListTail
	: Nothing | ',' DWordConstExpr DWordListTail
	;


PackageTerm
	: TOK_PACKAGE
	  '(' PackageNumElements ')'
	  '{' PackageList '}'
	;
PackageNumElements
	: Nothing | ByteConstExpr
	;


PackageList
	: Nothing | PackageElement PackageListTail
	;
PackageListTail
	: Nothing | ',' PackageElement PackageListTail
	| ',' Nothing /* XXX */
	;
PackageElement
	: DataObject | NameString
	;


EISAIDTerm
	: TOK_EISAID
	  '(' StringData ')'
	;


ResourceTemplateTerm
	: TOK_RESOURCETEMPLATE '(' ')'
	  '{' ResourceMacroList '}'
	;


UnicodeTerm
	: TOK_UNICODE
	  '(' StringData ')'
	;


ResourceMacroList
	: Nothing | ResourceMacroTerm ResourceMacroList
	;
ResourceMacroTerm
	: DMATerm | DWordIOTerm | DWordMemoryTerm
	| EndDependentFnTerm | FixedIOTerm | InterruptTerm
	| IOTerm | IRQNoFlagsTerm | IRQTerm | Memory24Term
	| Memory32FixedTerm | Memory32Term | QWordIOTerm
	| QWordMemoryTerm | RegisterTerm | StartDependentFnTerm
	| StartDependentFnNoPriTerm | VendorLongTerm
	| VendorShortTerm | WordBusNumberTerm | WordIOTerm
	;


DMATerm
	: TOK_DMA
	  '(' DMATypeKeyword ',' BusMasterKeyword ',' XferTypeKeyword ',' OptionalResourceTag ')'
	  '{' ByteList '}'
	;
OptionalResourceTag
	: Nothing | NameString
	;


DWordIOTerm
	: TOK_DWORDIO
	  '(' OptionalResourceType ',' OptionalMinType ',' OptionalMaxType ','
	  OptionalDecode ',' OptionalRangeType ',' DWordConstExpr ','
	  DWordConstExpr ',' DWordConstExpr ',' DWordConstExpr ','
	  DWordConstExpr ',' OptionalResSourceIndex ',' OptionalResSource ','
	  OptionalResourceTag ',' OptionalType ',' OptionalTranslationType ')'
	;
OptionalResourceType
	: Nothing | ResourceTypeKeyword
	;
OptionalMinType
	: Nothing | MinKeyword
	;
OptionalMaxType
	: Nothing | MaxKeyword
	;
OptionalDecode
	: Nothing | DecodeKeyword
	;
OptionalRangeType
	: Nothing | RangeTypeKeyword
	;
OptionalResSourceIndex
	: Nothing | ByteConstExpr
	;
OptionalResSource
	: Nothing | StringData
	;
OptionalType
	: Nothing | TypeKeyword
	;
OptionalTranslationType
	: Nothing | TranslationKeyword
	;


DWordMemoryTerm
	: TOK_DWORDMEMORY
	  '(' OptionalResourceType ',' OptionalDecode ',' OptionalMinType ','
	  OptionalMaxType ',' OptionalMemType ',' ReadWriteKeyword ','
	  DWordConstExpr ',' DWordConstExpr ',' DWordConstExpr ','
	  DWordConstExpr ',' DWordConstExpr ',' OptionalResSourceIndex ','
	  OptionalResSource ',' OptionalResourceTag ',' OptionalAddressRange ','
	  OptionalTranslationType ')'
	;
OptionalMemType
	: Nothing | MemTypeKeyword
	;
OptionalAddressRange
	: Nothing | AddressKeyword
	;


EndDependentFnTerm
	: TOK_ENDDEPENDENTFN '(' ')'
	;


FixedIOTerm
	: TOK_FIXEDIO
	  '(' WordConstExpr ',' ByteConstExpr ',' OptionalResourceTag ')'
	;


InterruptTerm
	: TOK_INTERRUPT
	  '(' OptionalResourceType ',' InterruptTypeKeyword ','
	  InterruptLevelKeyword ',' OptionalShareType ','
	  OptionalResSourceIndex ',' OptionalResSource ','
	  OptionalResourceTag ')'
	  '{' DWordList '}'
	;
OptionalShareType
	: Nothing | ShareTypeKeyword
	;


IOTerm
	: TOK_IO
	  '(' IODecodeKeyword ',' WordConstExpr ',' WordConstExpr ','
	  ByteConstExpr ',' ByteConstExpr ',' OptionalResourceTag ')'
	;


IRQNoFlagsTerm
	: TOK_IRQNOFLAGS
	  '(' OptionalResourceTag ')'
	  '{' ByteList '}'
	;


IRQTerm
	: TOK_IRQ
	  '(' InterruptTypeKeyword ',' InterruptLevelKeyword ',' 
	  OptionalShareType ',' OptionalResourceTag ')'
	  '{' ByteList '}'
	;


Memory24Term
	: TOK_MEMORY24
	  '(' ReadWriteKeyword ',' WordConstExpr ','
	  WordConstExpr ',' WordConstExpr ',' WordConstExpr ','
	  OptionalResourceTag ')'
	;


Memory32FixedTerm
	: TOK_MEMORY32FIXED
	  '(' ReadWriteKeyword ',' DWordConstExpr ',' DWordConstExpr ','
	  OptionalResourceTag ')'
	;


Memory32Term
	: TOK_MEMORY32
	  '(' ReadWriteKeyword ',' DWordConstExpr ',' DWordConstExpr ','
	  DWordConstExpr ',' DWordConstExpr ',' OptionalResourceTag ')'
	;

QWordIOTerm
	: TOK_QWORDIO
	  '(' OptionalResourceType ',' OptionalMinType ',' OptionalMaxType ','
	  OptionalDecode ',' OptionalRangeType ',' QWordConstExpr ','
	  QWordConstExpr ',' QWordConstExpr ',' QWordConstExpr ','
	  QWordConstExpr ',' OptionalResSourceIndex ',' OptionalResSource ','
	  OptionalResourceTag OptionalType OptionalTranslationType ')'
	;


QWordMemoryTerm
	: TOK_QWORDMEMORY
	  '(' OptionalResourceType ',' OptionalDecode ',' OptionalMinType ','
	  OptionalMaxType ',' OptionalMemType ',' ReadWriteKeyword ','
	  QWordConstExpr ',' QWordConstExpr ',' QWordConstExpr ','
	  QWordConstExpr ',' QWordConstExpr ',' OptionalResSourceIndex ','
	  OptionalResSource ',' OptionalResourceTag OptionalAddressRange
	  OptionalTranslationType ')'
	;

RegisterTerm
	: TOK_REGISTER
	  '(' AddressSpaceKeyword ',' ByteConstExpr  ',' ByteConstExpr  ','
	  QWordConstExpr ')'
	;


StartDependentFnTerm
	: TOK_STARTDEPENDENTFN
	  '(' ByteConstExpr ',' ByteConstExpr ')'
	  '{' ResourceMacroList '}'
	;


StartDependentFnNoPriTerm
	: TOK_STARTDEPENDENTFNNOPRI '(' ')'
	  '{' ResourceMacroList '}'
	;


VendorLongTerm
	: TOK_VENDORLONG
	  '(' OptionalResourceTag ')'
	  '{' ByteList '}'
	;


VendorShortTerm
	: TOK_VENDORSHORT
	  '(' OptionalResourceTag ')'
	  '{' ByteList '}'
	;


WordBusNumberTerm
	: TOK_WORDBUSNUMBER
	  '(' OptionalResourceType ',' OptionalMinType ',' OptionalMaxType ','
	  OptionalDecode ',' WordConstExpr ',' WordConstExpr ','
	  WordConstExpr ',' WordConstExpr ',' WordConstExpr ','
	  OptionalResSourceIndex ',' OptionalResSource ',' OptionalResourceTag ')'
	;


WordIOTerm
	: TOK_WORDIO
	  '(' OptionalResourceType ',' OptionalMinType ',' OptionalMaxType ','
	  OptionalDecode ',' OptionalRangeType ',' WordConstExpr ','
	  WordConstExpr ',' WordConstExpr ',' WordConstExpr ',' WordConstExpr ','
	  OptionalResSourceIndex ',' OptionalResSource ',' OptionalResourceTag
	  OptionalType OptionalTranslationType ')'
	;

/*
 * ASL Names
 */

RootChar
	: '\\'
	;
NameSeg
	: TOK_NAMESEG
	;
NameString
	: RootChar NamePath
		{
			$$ = parser_strcat("\\", $2);
			free($2);
			goto namestring_out;
		}
	| '^' PrefixPath NamePath
		{
			char	*ptr;
			ptr = parser_strcat($2, $3);
			$$ = parser_strcat("^", ptr);
			free(ptr);
			free($2);
			free($3);
			goto namestring_out;
		}
	| NonEmptyNamePath
		{
			struct asl_symtree	*ptr;

			$$ = $1;
		namestring_out:
			ptr = symtree_new(SYMTYPE_NameString, $$);
			ptr->symbol.NameString.ptr = $$;
		}
	;
PrefixPath
	: Nothing		{ $$ = NULL; }
	| '^' PrefixPath	{ $$ = parser_strcat("^", $2); free($2); }
	;
NamePath
	: Nothing		{ $$ = NULL; }
	| NameSeg NamePathTail	{ $$ = parser_strcat($1, $2); free($1); free($2); }
	;
NonEmptyNamePath
	: NameSeg		{ $$ = $1; }
	| NameSeg NamePathTail	{ $$ = parser_strcat($1, $2); free($1); free($2); }
	;
NamePathTail
	: Nothing		{ $$ = NULL; }
	| '.' NameSeg NamePathTail
		{
			char	*ptr;
			ptr = parser_strcat($2, $3);
			$$ = parser_strcat(".", ptr);
			free(ptr);
			free($2);
			free($3);
		}
	;

/*
 * Nothing
 */

Nothing
	: /* empty */
	;

%%

static char *
parser_strcat(const char *str1, const char *str2)
{
	char	*str;
	size_t	len;

	len = ((str1 == NULL) ? 0 : strlen(str1)) +
	      ((str2 == NULL) ? 0 : strlen(str2)) + 1;

	if ((str = (char *)malloc(len)) == NULL) {
		parser_fatal("No more room for strcat");
	}

	bzero(str, len);
	if (str1 != NULL) {
		strncpy(str, str1, strlen(str1));
	}
	if (str2 != NULL) {
		strcat(str, str2);
	}

	return(str);
}

struct parser_scope {
	TAILQ_ENTRY(parser_scope)	links;
	char			*name;
};
TAILQ_HEAD(, parser_scope)	scope_list;

static __inline void
parser_scope_init(void)
{
	static int	first = 1;

	if (first) {
		TAILQ_INIT(&scope_list);
		first = 0;
	}
}

static void
parser_scope_enter(const char *name)
{
	struct parser_scope *new_entry;

	parser_scope_init();

	if ((new_entry = malloc(sizeof(struct parser_scope))) == NULL) {
		parser_fatal("No more room for parser_scope.");
	}

	bzero(new_entry, sizeof(struct parser_scope));
	new_entry->name = strdup(name);
	TAILQ_INSERT_HEAD(&scope_list, new_entry, links);
}

static void
parser_scope_exit(void)
{
	struct parser_scope *entry;

	parser_scope_init();
	entry = TAILQ_FIRST(&scope_list);
	TAILQ_REMOVE(&scope_list, entry, links);
	free(entry->name);
	free(entry);
}

static char *
parser_scope_fullpath(const char *name)
{
	struct parser_scope *entry;
	static char	fullpath[PARSER_MSGBUFSIZE];

	parser_scope_init();
	bzero(fullpath, sizeof(fullpath));
	TAILQ_FOREACH(entry, &scope_list, links) {
		snprintf(fullpath, sizeof(fullpath), "%s%s.", fullpath, entry->name);
	}
	snprintf(fullpath, sizeof(fullpath), "%s%s", fullpath, name);
	return(fullpath);
}

void
yyerror(const char *msg)
{

	fprintf(stderr, "line %d: %s\n", parser_lineno, msg);
}

#define PARSER_WARN(msg)	do {					\
	va_list	ap;							\
	va_start(ap, msg);						\
	vsnprintf(parser_msgbuf, PARSER_MSGBUFSIZE, msg, ap);		\
	va_end(ap);							\
	yyerror(parser_msgbuf);						\
} while (0)

void
parser_warn(const char *msg, ...)
{
	PARSER_WARN(msg);
}

void
parser_error(const char *msg, ...)
{
	PARSER_WARN(msg);
	exit(1);
}

void
parser_fatal(const char *msg, ...)
{
	PARSER_WARN(msg);
	abort();
}

extern FILE	*yyin;

void
parser_main(const char *asl_filename)
{
	int	bytes;

	parser_phase = 1;
	parser_lineno = 1;
	if ((yyin = fopen(asl_filename, "r")) == NULL) {
		(void) err(1, "cannot open ASL file");
	}

	if (parser_debuglevel) {
		printf("==== PARSER PHASE %d ====\n", parser_phase);
	}
		
	if (yyparse() != 0) {
		parser_error("ASL parser errors in phase 1\n");
	}
	fclose(yyin);

	if (parser_debuglevel) {
		symtab_dump();
		symtree_dump(symtree_getroot(), 0);
	}

	parser_phase++;
	if (parser_debuglevel) {
		printf("==== PARSER PHASE %d ====\n", parser_phase);
	}
	generator_main();
	if (parser_debuglevel) {
		symtree_dump(symtree_getroot(), 0);
		generator_dump_amlcode();
	}

	parser_phase++;
	if (parser_debuglevel) {
		printf("==== PARSER PHASE %d ====\n", parser_phase);
	}
	bytes = generator_create_amlfile(aml_filename);
	if (parser_debuglevel) {
		printf("%d bytes written\n", bytes);
	}

	symtab_destroy();
	symtree_purge(symtree_getroot()->child);
}
