/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aslc.h"

TAILQ_HEAD(, asl_symtab)	symtab_list;

static __inline void
symtab_init(void)
{
	static int	first = 1;

	if (first) {
		TAILQ_INIT(&symtab_list);
		first = 0;
	}
}

struct asl_symtab *
symtab_new(const char *name, int type)
{
	struct asl_symtab *new_entry;

	symtab_init();
	if ((new_entry = malloc(sizeof(struct asl_symtab))) == NULL) {
		parser_fatal("No more room for symbols.");
	}

	bzero(new_entry, sizeof(struct asl_symtab));
	new_entry->name = strdup(name);
	new_entry->type = type;
	TAILQ_INSERT_TAIL(&symtab_list, new_entry, links);
	return(new_entry);
}

struct asl_symtab *
symtab_find(const char *name)
{
	struct asl_symtab *entry;

	symtab_init();
	TAILQ_FOREACH(entry, &symtab_list, links) {
		if (strcmp(entry->name, name) == 0) {
			return(entry);
		}
	}
	return(NULL);
}

void
symtab_dump(void)
{
	struct asl_symtab *entry;

	symtab_init();
	if (parser_debuglevel == 0) {
		return;
	}

	TAILQ_FOREACH(entry, &symtab_list, links) {
		printf("SYMTAB: %s(%d)\n", entry->name, entry->type);
	}
}

void
symtab_destroy(void)
{
	struct asl_symtab *entry;

	symtab_init();
	while ((entry = TAILQ_FIRST(&symtab_list))) {
		TAILQ_REMOVE(&symtab_list, entry, links);
		free(entry->name);
		free(entry);
	}
}

const char *
symtab_getname(struct asl_symtab *sym)
{
	return(sym->name);
}

void
symtab_method_setflags(struct asl_symtab *sym, int flags)
{
	if (sym->type == SYM_METHOD) {
		sym->data.method.flags = (u_int8_t)(flags & 0xff);
	}
}

u_int8_t
symtab_method_getnumargs(struct asl_symtab *sym)
{
	if (sym->type == SYM_METHOD) {
		return(sym->data.method.flags & 7);
	}
	return(0);
}
