/*-
 * ACPI Source Language Compiler
 *
 * Copyright (c) 2000 Mitsuru IWASAKI <iwasaki@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "aslc.h"
#include "aslcsymtree.h"

static struct asl_symtree	 symtree_root;
static struct asl_symtree	*symtree_curptr = &symtree_root;

static __inline void
symtree_init(void)
{
	static int	first = 1;

	if (first) {
		bzero(&symtree_root, sizeof(symtree_root));
		symtree_root.symbol.common.name = strdup("ASLCode");
		first = 0;
	}
}

struct asl_symtree *
symtree_new(int type, const char *symbol)
{
	struct asl_symtree *new_entry, *brother;

	symtree_init();
	if ((new_entry = malloc(sizeof(struct asl_symtree))) == NULL) {
		parser_fatal("No more room for symbols.");
	}

	bzero(new_entry, sizeof(struct asl_symtree));
	new_entry->symbol.common.type = type;
	new_entry->symbol.common.lineno = parser_lineno;
	new_entry->symbol.common.name = strdup(symbol);
	new_entry->parent = symtree_curptr;
	new_entry->child = NULL;
	if (symtree_curptr->child != NULL) {
		brother = symtree_curptr->child;
		while (brother) {
			if (brother->brother == NULL) {
				brother->brother  = new_entry;
				break;
			}
			brother = brother->brother;
		}
	} else {
		symtree_curptr->child = new_entry;
	}

	return(new_entry);
}

void
symtree_moveptr(struct asl_symtree *symbol)
{
	symtree_init();
	if (symbol != NULL) {
		symtree_curptr = symbol;
	}
}

void
symtree_upptr(void)
{
	symtree_init();
	symtree_moveptr(symtree_curptr->parent);
}

void
symtree_purge(struct asl_symtree *start)
{
	symtree_init();
	if (start != NULL) {
		if (start->child != NULL) {
			symtree_purge(start->child);
		}

		if (start->brother != NULL) {
			symtree_purge(start->brother);
		}
		start->parent->child = NULL;
		free(start->symbol.common.name);
		free(start);
	} else {
		symtree_purge(symtree_curptr->child);
	}
}

struct asl_symtree *
symtree_getptr(void)
{
	symtree_init();
	return(symtree_curptr);
}

struct asl_symtree *
symtree_getroot(void)
{
	symtree_init();
	return(&symtree_root);
}

void
symtree_dump(struct asl_symtree *start, int depth)
{
	int	i;

	symtree_init();
	for (i = 0; i < depth; i++) {
		printf(" ");
	}
	printf("<%s:0x%x:0x%x>\tat line %d\n", start->symbol.common.name,
	    start->symbol.common.encoding, start->symbol.common.pkglength,
	    start->symbol.common.lineno);
	if (start->child != NULL) {
		symtree_dump(start->child, depth + 1);
	}

	if (start->brother != NULL) {
		symtree_dump(start->brother, depth);
	}
}
